# LeNet5 from Paperspace blog https://blog.paperspace.com/writing-lenet5-from-scratch-in-python/
import torch.nn as nn

#Defining the convolutional neural network
class LeNet5(nn.Module):
    def __init__(self, num_classes):
        super(LeNet5, self).__init__()
        self.layer1 = nn.Sequential(
            nn.Conv2d(1, 6, kernel_size=5, stride=1, padding=0),
            nn.BatchNorm2d(6),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size = 2, stride = 2))
        self.layer2 = nn.Sequential(
            nn.Conv2d(6, 16, kernel_size=5, stride=1, padding=0),
            nn.BatchNorm2d(16),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size = 2, stride = 2))
        self.fc = nn.Linear(400, 120)
        self.relu = nn.ReLU()
        self.fc1 = nn.Linear(120, 84)
        self.relu1 = nn.ReLU()
        self.fc2 = nn.Linear(84, num_classes)
        
    def forward(self, x):
        out = self.layer1(x)
        out = self.layer2(out)
        out = out.reshape(out.size(0), -1)
        out = self.fc(out)
        out = self.relu(out)
        out = self.fc1(out)
        out = self.relu1(out)
        out = self.fc2(out)
        return out


from mp_api.client import MPRester
from pymatgen.analysis.diffraction.xrd import XRDCalculator
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

# from mendeleev import element
# elements = {}
# for i in range(1,119):
# 	elements[element(i).symbol] = i
elements={'H': 1, 'He': 2, 'Li': 3, 'Be': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10, 'Na': 11, 'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15, 'S': 16, 'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20, 'Sc': 21, 'Ti': 22, 'V': 23, 'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29, 'Zn': 30, 'Ga': 31, 'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39, 'Zr': 40, 'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44, 'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49, 'Sn': 50, 'Sb': 51, 'Te': 52, 'I': 53, 'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59, 'Nd': 60, 'Pm': 61, 'Sm': 62, 'Eu': 63, 'Gd': 64, 'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70, 'Lu': 71, 'Hf': 72, 'Ta': 73, 'W': 74, 'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79, 'Hg': 80, 'Tl': 81, 'Pb': 82, 'Bi': 83, 'Po': 84, 'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89, 'Th': 90, 'Pa': 91, 'U': 92, 'Np': 93, 'Pu': 94, 'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99, 'Fm': 100, 'Md': 101, 'No': 102, 'Lr': 103, 'Rf': 104, 'Db': 105, 'Sg': 106, 'Bh': 107, 'Hs': 108, 'Mt': 109, 'Ds': 110, 'Rg': 111, 'Cn': 112, 'Nh': 113, 'Fl': 114, 'Mc': 115, 'Lv': 116, 'Ts': 117, 'Og': 118}
 
from mlbands.misc import *

import numpy as np
import matplotlib.pyplot as plt

class Material:
    def __init__( self, api_key, material_ID = 'mp-1103503', box_array = None):
        '''Material class. To store and process material data

        Parameters
        ----------
        API_KEY : str
            secret API KEY used to run MPRester data requests
        material_ID : int
            ID of selected material (without "mp-" prefix)
        box_array : (float, float, float)
            3-D Tensor representing material chemical env. 
        '''
        self.API_KEY = api_key
        self.material_ID = material_ID
        self.box_array = box_array          # material already transformed to box data form (optional)

    def bands(self,nonzero_gap=False):
        '''obtain band gap information for requested Material 

        Parameters
        ----------
        nonzero_gap: bool
            if True, only returns nonzero band gaps 
        '''
        with MPRester(api_key=self.API_KEY) as mpr:
            #adapted from https://matsci.org/t/obtain-large-numbers-of-band-structures/3780
            bandstructure = None 
            try:
                bandstructure = mpr.get_bandstructure_by_material_id(self.material_ID,line_mode=False)
            except:
                pass
            if bandstructure:
                band_gap = bandstructure.get_band_gap()
                print('Band Gap: {} eV\nDirect Gap: {}\nMetallic: {}'.\
                    format(band_gap['energy'],\
                        'Yes' if band_gap['direct'] else 'No',\
                        'No' if band_gap['transition'] else 'Yes'))
                        
                if nonzero_gap:
                    if band_gap['energy']: return band_gap
                    else: return 0
                else: return band_gap
            else:
                return 0

    def load_structure(self, conventional=True):
        '''helper function to load structural information of Material

        Parameters
        ----------
        conventional: bool
            if True, returns conventional standard structure else returns primitive
        '''
        with MPRester(self.API_KEY) as mpr:

            # first retrieve the relevant structure
            structure = mpr.get_structure_by_material_id(self.material_ID)
        
        # important to use the conventional structure to ensure
        # that peaks are labelled with the conventional Miller indices
        sga = SpacegroupAnalyzer(structure)
        conventional_structure = sga.get_conventional_standard_structure()
        
        if conventional:
            structure = conventional_structure

        return structure

       
    def structural(self):
        '''prints structural information
        '''
        structure = Material(self.API_KEY, self.material_ID).load_structure()

        print(structure.lattice)
        print(structure.sites)  #https://pymatgen.org/pymatgen.core.sites.html?highlight=periodicsite#pymatgen.core.sites.PeriodicSite

        Nsites = len(structure.sites)
        for i in range(Nsites):
            print('\n\n')
            # print(structure.sites[i])
            print(structure.sites[i].species)
            print(structure.sites[i].coords)
            print(structure.sites[i].frac_coords)     
    

    def to_xyz(self, fractional=False):
        '''processes the material structure as an array with an xyz coordinate format 
        i.e. [atom_numbers, x_coord, y_coord, z_coord]

        Parameters
        ----------
        fractional: bool
            if False, returns xyz coordinates else returns fractional (a,b,c) coordinates
        '''
        structure = Material(self.API_KEY, self.material_ID).load_structure(conventional=True)

        Nsites = len(structure.sites)

        xyz_array = np.zeros((Nsites,4))

        if not fractional:
            for i in range(Nsites):
                # atom_num = elements(str(structure.sites[i].specie)).atomic_number # specie is not a typo
                atom_num = elements[str(structure.sites[i].specie)] # specie is not a typo

                xyz_array[i] = [atom_num,*structure.sites[i].coords]
        else:
            for i in range(Nsites):
                # atom_num = elements(str(structure.sites[i].specie)).atomic_number 
                atom_num = elements[str(structure.sites[i].specie)] 

                xyz_array[i] = [atom_num,*structure.sites[i].frac_coords]

        return xyz_array


    def to_box(self, fractional=False):
        '''processes the material structure as a sparse, 3-D Tensor "Box" with atom_numbers as the 
        non-zero entries
        i.e. box = Tensor((z_coord, y_coord, x_coord))

        Parameters
        ----------
        fractional: bool
            if False, returns tensor space from xyz coordinates else from fractional (a,b,c) coordinates
        '''
        xyz_array = Material(self.API_KEY, self.material_ID).to_xyz(fractional)

        coords = xyz_array[:,1:]            
        MAX = np.ceil(np.max(coords)).astype('int')
        MIN = np.floor(np.min(coords)).astype('int')
        
        xyz_array[:,1:] -= MIN

        L = (MAX-MIN+1)

        box = np.zeros((L,L,L))
        for i in xyz_array:
            print(i)

            atom,x,y,z = i.astype('int')
            box[z,y,x] = atom

        return box
            
    def visual(self, spacing = 1, fractional=False):
        '''simple visualization tool for Material (plot)

        Parameters
        ----------
        spacing: float
            increases spacing by multiplying coordinate distance
        fractional: bool
            if False, returns tensor space from xyz coordinates else from fractional (a,b,c) coordinates
        '''
        ax = plt.axes(projection='3d')
        colors = np.linspace(2**20,2**24,118,dtype='int') #divide color range into 118 colors (for the 118 chemical elements)


        if self.box_array is not None:
            #presence of box_array supplants material_ID
            for i in np.argwhere(self.box_array):
                x,y,z = i
                atom = int(self.box_array[tuple(i)])
                ax.scatter3D(x,y,z, s=100, c="#"+hex(colors[atom])[2:])

        else:
            xyz_array = Material(self.API_KEY, self.material_ID).to_xyz(fractional)
            xyz_array[:,1:]*=spacing

            for i in xyz_array:
                atom,*xyz = i.astype('int')
                ax.scatter3D(*xyz, s=100, c="#"+hex(colors[atom])[2:])


        set_axes_equal(ax)           

        plt.axis('off')
        plt.show()


    def XRD(self):
        '''XRD pattern data for Material
        '''
        structure = Material(self.API_KEY, self.material_ID).load_structure(self.API_KEY)
        
        # this example shows how to obtain an XRD diffraction pattern
        # these patterns are calculated on-the-fly from the structure
        calculator = XRDCalculator(wavelength='CuKa')
        pattern = calculator.get_pattern(structure)

        print('\npattern:\n',pattern)
        

    def thermo(self):   
        '''thermodynamic data for Material
        '''
        with MPRester(api_key=self.API_KEY) as mpr:

            # for a single material
            # thermo_docs = mpr.thermo.get_data_by_id(self.material_ID)      #   DOES NOT WORK msg(Item with thermo_id = mp-1103503 not found)
            
            # for many materials, it's much faster to use
            # the `search` method, where additional material_ids can 
            # be added to this list
            thermo_docs = mpr.thermo.search(material_ids=[self.material_ID])

        # print(thermo_docs[0].energy_per_atom)
        return thermo_docs

    def magnetism(self):      
        '''magnetic properties data for Material
        '''
        with MPRester(api_key=self.API_KEY) as mpr:
            magnetism_doc = mpr.magnetism.get_data_by_id('mp-1103503')

        # print(magnetism_doc.total_magnetization)
        return magnetism_doc

class Group:
    def __init__(self,api_key):
        '''Group class for bulk calculations

        Parameters
        ----------
        API_KEY : str
            secret API KEY used to run MPRester data requests
        materials : int
            IDs of selected material (with "mp-" prefix)
        X : [varies] (float,float,float); ((float,float,float), *float); (*float)
            quantitative predictors ("X-value")
        Y : float
            band gap (eV) quantitative response ("Y-value","label")
        '''
        self.API_KEY = api_key
        self.materials = []         # material IDs (materials found in ID_list; see data_make inputs)
        self.X = []                 # quantitative predictors ("X-value")
        self.Y = []                 # quantitative response ("Y-value","label")
        
        self.box_lengths = []       # length scale of boxes (3-D Tensors of material chemical env.)
        self.max_length = 0

    def transfer(self, loaded_data):
        '''transfer constructor variable information to Group class from loaded data

        Parameters
        ----------
        loaded data : <"Group" class object>
            file containing processed Materials in a Group container
        '''
        self.X = loaded_data.X
        self.Y = loaded_data.Y
        self.materials = loaded_data.materials

    def data_make(self, ID_list = range(1,10), nonzero_gap=False ):    
        '''generate Group data for selected ID_list numbers [ for materials which exist with such IDs ]

        Parameters
        ----------
        ID_list : list[int]
            list of material_ID numbers (i.e. without "mp-" prefix)
        nonzero_gap: bool
            if True, only returns nonzero band gaps 
        '''
        for i in ID_list:
            material = Material(self.API_KEY, 'mp-'+str(i))
            BG = material.bands(nonzero_gap)
            if BG:                              # if material with material_ID exists
                self.materials.append('mp-'+str(i))

                # quantitative response ("Y-value","label")
                self.Y.append(BG['energy'])     # append band gap energy ( eV ) 

                box = material.to_box()
                self.box_lengths.append(box.shape[0])
                # quantitative predictors ("X-value")
                self.X.append(box)              # append 3-D Tensor representing material chemical env. 
                # for extra_x in features:
                #     self.X.append(extra_x)

        self.Y = np.array(self.Y)

    def data_expand(self,boxes=True,*property_funcs):
        '''expand "X-values" data to other chemical characteristics 

        e.g. data_expand(True, thermo, magnetic) appends "thermo" and "magnetic" characteristics to X in addition to 3-D Tensors
        e.g. data_expand(False, thermo, magnetic) creates a new X-value object and appends "thermo" and "magnetic" characteristics alone

        Parameters
        ----------
        boxes : bool
            removes 3-D tensors to X variable if False, 
        *property_funcs: <function_1>, ..., <function_N>
            operates functions on Materials which extract additional properties and appends them to X
        '''
        boxes = self.X
        self.X = [] 
        # load a property for all materials
        for i in self.materials:
            material = Material(self.API_KEY, 'mp-'+str(i))

            k=0
            material_props = []
            material_props.append(boxes[k]) if boxes else None
            for func in property_funcs:
                material_props.append(material.func)
            
            self.X.append(material_props)
            k+=1

       
    def resize_boxes(self, L=32):
        '''resize 3-D tensor boxes to fit neural network architecture (32x32x32)

        Parameters
        ----------
        L: int
            length-scale of boxes (3-D Tensors representing material chemical env.)
        '''
        max_length = np.max(self.box_lengths)

        if L >= max_length:


            self.X = np.array([  np.pad(self.X[i],\
                        ( (0,L-self.box_lengths[i]),(0,L-self.box_lengths[i]),(0,L-self.box_lengths[i]) ) 
                        ) for i in range(len(self.X)) ])

        # print(self.X.shape)

    





# Load in relevant libraries, and alias where appropriate
import torch
import torch.nn as nn
import torchvision
import torchvision.transforms as transforms

from torch.utils.data import Dataset, DataLoader
class Data(Dataset):
    def __init__(self,X_train,Y_train):
        self.X=torch.from_numpy(X_train).float()
        self.Y=torch.from_numpy(Y_train).float()
        self.len=self.X.shape[0]
    def __getitem__(self,index):      
        return self.X[index], self.Y[index]
    def __len__(self):
        return self.len

from mlbands.neuralnets import LeNet3D, LeNet5


def reshapeX(array,channels=1):
    return array.reshape((array.shape[0],1,*array.shape[1:]))

def reshapeY(array):
    if len(array.shape)==1:
        return array.reshape(-1,1)
    else: 
        return array

def reshapeXY(data,channels=1):

    X,Y = data
    # return [reshapeX(X,channels=1),Y]
    return [reshapeX(X,channels=1),reshapeY(Y)]


# Device will determine whether to run the training on GPU or CPU.
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')


class Machine:
    def __init__(self, batch_size=1, num_classes=1, learning_rate=0.001, num_epochs=10 ):
        '''Machine [ learning ] class for bulk calculations

        Parameters
        ----------
        batch_size  : int
            batch size
        num_classes : int
            number of output ("Y") classes
        learning_rate : float
            learning rate for optimizer
        num_epochs : int
            number of training epochs
        neuralnet: < neuralnets.neural_network Class>
            neural network model (default: LeNet3D)
        cost : <torch.nn Object>
            cost / loss function for training default: MSELoss
        input_channels : int
            number of input channels to neural network for (3-D structure, band_gap) (X,Y) combination, it is =1
        '''
        self.batch_size = batch_size 
        self.num_classes = num_classes 
        self.learning_rate = learning_rate
        self.num_epochs = num_epochs

        self.neuralnet = LeNet3D
        self.cost = nn.MSELoss()            # the loss function
        # self.cost = nn.CrossEntropyLoss()
        self.input_channels = 1


    def learn(self, trainset, testset):
        '''neural network training and testing (validation)

        Parameters
        ----------
        trainset : < Group.X, Group.y >
            training set of x and y values
        testset : < Group.X, Group.y >
            testing set of x and y values
        '''
        # Train the model
        trainset = reshapeXY(trainset,channels=self.input_channels)
        train_dataset=Data(*trainset)

        train_loader = torch.utils.data.DataLoader(dataset = train_dataset,
                                                batch_size = self.batch_size,
                                                shuffle = True)

        model = self.neuralnet(self.num_classes).to(device)


        #Setting the optimizer with the model parameters and learning rate
        optimizer = torch.optim.Adam(model.parameters(), lr=self.learning_rate)

        #this is defined to print how many steps are remaining when training
        total_step = len(train_loader)


        # Training the model
        for epoch in range(self.num_epochs):
            for i, (images, labels) in enumerate(train_loader):  
                images = images.to(device)
                labels = labels.to(device)
                
                #Forward pass
                outputs = model(images)
                loss = self.cost(outputs, labels)
                    
                # Backward and optimize
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                        
                if (i+1) % 400 == 0:
                    print ('Epoch [{}/{}], Step [{}/{}], Loss: {:.4f}' 
                                .format(epoch+1, self.num_epochs, i+1, total_step, loss.item()))


        # Test the model
        # In test phase, we don't need to compute gradients (for memory efficiency)

        testset = reshapeXY(testset,channels=self.input_channels)

        test_dataset=Data(*testset)

        test_loader = torch.utils.data.DataLoader(dataset = test_dataset,
                                        batch_size = self.batch_size,
                                        shuffle = True)


        with torch.no_grad():
            correct = 0
            total = 0
            for images, labels in test_loader:
                images = images.to(device)
                labels = labels.to(device)
                outputs = model(images)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

                print('predicted: {} ground-truth: {}'.format(predicted,labels ))

            print('Accuracy of the network on the test data: {} %'.format(100 * correct / total))




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mlbands', 'mlbands.neuralnets']

package_data = \
{'': ['*']}

install_requires = \
['mp-api>=0.30.5,<0.31.0',
 'torch>=1.13.1,<2.0.0',
 'torchvision>=0.14.1,<0.15.0']

setup_kwargs = {
    'name': 'mlbands',
    'version': '1.0.1',
    'description': 'A Python package that implements automatic prediction of electronic band gaps for a set of materials based on training data',
    'long_description': '# ML Band Gaps (Materials)\n\n> Ideal candidate: skilled ML data scientist with solid knowledge of materials science.\n\n# Overview\n\nThe aim of this task is to create a python package that implements automatic prediction of electronic band gaps for a set of materials based on training data.\n\n# User story\n\nAs a user of this software I can predict the value of an electronic band gap after passing training data and structural information about the target material.\n\n# Requirements\n\n- suggest the bandgap values for a set of materials designated by their crystallographic and stoichiometric properties\n- the code shall be written in a way that can facilitate easy addition of other characteristics extracted from simulations (forces, pressures, phonon frequencies etc)\n\n# Expectations\n\n- the code shall be able to suggest realistic values for slightly modified geometry sets - eg. trained on Si and Ge it should suggest the value of bandgap for Si49Ge51 to be between those of Si and Ge\n- modular and object-oriented implementation\n- commit early and often - at least once per 24 hours\n\n# Timeline\n\nWe leave exact timing to the candidate. Must fit Within 5 days total.\n\n# Notes\n\n- use a designated github repository for version control\n- suggested source of training data: materialsproject.org\n',
    'author': 'Andrew R. Garcia',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# https://github.com/andrewrgarcia/3D-LeNet-with-PyTorch
from torch.autograd import Variable
import torch
import torch.nn as nn
import torch.nn.functional as F

L=32    # length scale of cubes (default: 32 x 32 x 32)

class LeNet3D(nn.Module):
    def __init__(self, num_classes):
        super(LeNet3D, self).__init__()

        self.conv1 = nn.Conv3d(1, 6, kernel_size=(5, 5, 5))
        self.pool = nn.MaxPool3d(2, 2)
        self.conv2 = nn.Conv3d(6, (L//2), kernel_size=(5, 5, 5))
        self.fc1 = nn.Linear((L//2) * 5 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, num_classes)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        # print(x.size())
        x = self.pool(F.relu(self.conv2(x)))
        # print(x.size())
        x = x.view(-1, (L//2) * 5 * 5 * 5)
        # print(x.size())
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


# # input: (N = batch_size, C = 1, L = 32, L = 32, L = 32)
# # output: (N, num_classes)

# # Test the model:
# num_classes = 1

# model = LeNet3D(num_classes)

# x = Variable(torch.randn(10, 1, L, L, L)) # (N_samples,C_channels,D=L,H=L,W=L)
# print(x)
# y = model(x)
# print(y)






import numpy as np

def set_axes_radius(ax, origin, radius):
    '''set_axes_radius and set_axes_equal * * * Credit:
    Mateen Ulhaq (answered Jun 3 '18 at 7:55)
    https://stackoverflow.com/questions/13685386/matplotlib-equal-unit-length-with-equal-aspect-ratio-z-axis-is-not-equal-to'''
    ax.set_xlim3d([origin[0] - radius, origin[0] + radius])
    ax.set_ylim3d([origin[1] - radius, origin[1] + radius])
    ax.set_zlim3d([origin[2] - radius, origin[2] + radius])
def set_axes_equal(ax):
    '''Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.
    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    '''
    limits = np.array([
        ax.get_xlim3d(),
        ax.get_ylim3d(),
        ax.get_zlim3d(),
    ])

    origin = np.mean(limits, axis=1)
    radius = 0.5 * np.max(np.abs(limits[:, 1] - limits[:, 0]))
    set_axes_radius(ax, origin, radius)


# adapted from https://stackoverflow.com/a/47381855/14460178
import pickle

def save(data, filename = 'file.mat3r'): 
    with open(filename, 'wb') as handle: pickle.dump(data, handle)

def load(filename = 'file.mat3r'): 
    with open(filename, 'rb') as handle: return pickle.load(handle)
